
/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 * Low-level routines
 *
 * Copyright (C) 1998-2002 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <asm/io.h>

#include "catweasel.h"

static __inline__ void CWSetCReg(catweasel_contr *c, unsigned char clear, unsigned char set)
{
    c->control_register = (c->control_register & ~clear) | set;
    outb(c->control_register, c->iobase + 2);
}

static void CWTriggerStep(catweasel_contr *c)
{
    CWSetCReg(c, 1<<0, 0);
    c->msdelay(3);
    CWSetCReg(c, 0, 1<<0);
    c->msdelay(3);
}

void catweasel_init_controller(catweasel_contr *c)
{
    int i, j;

    if(!c->iobase)
	return;

    c->control_register = 255;

    /* select all drives, step inside */
    CWSetCReg(c, 1<<1 | 1<<4 | 1<<5, 0);
    for(i=0;i<3;i++)
	CWTriggerStep(c);

    for(i=0;i<2;i++) {
	c->drives[i].number = i;
	c->drives[i].contr = c;
	c->drives[i].diskindrive = 0;
	
	/* select only the respective drive, step to track 0 */
	CWSetCReg(c, 0, 1<<4 | 1<<5 | 1<<1);
	CWSetCReg(c, 1<<(5-i), 0);
	
	for(j = 0; j < 86 && inb(c->iobase+2) & 16; j++)
	    CWTriggerStep(c);
	
	if(j < 86) {
	    c->drives[i].type = 1;
	    c->drives[i].track = 0;
	} else {
	    c->drives[i].type = 0;
	}
    }
    
    CWSetCReg(c, 0, 1<<4 | 1<<5); /* deselect all drives */
}

void catweasel_free_controller(catweasel_contr *c)
{
    if(!c->iobase)
	return;

    /* all motors off, deselect all drives */
    CWSetCReg(c, 0, 1<<3 | 1<<7 | 1<<4 | 1<<5);
}

void catweasel_select(catweasel_contr *c, int dr0, int dr1)
{
    CWSetCReg(c, dr0 << 5 | dr1 << 4, (!dr0) << 5 | (!dr1) << 4);
}

static const unsigned char motorbits[2] = { 3, 7 };

void catweasel_set_motor(catweasel_drive *d, int on)
{
    int mask = 1 << motorbits[d->number];
    
    if(on)
	CWSetCReg(d->contr, mask, 0);
    else
	CWSetCReg(d->contr, 0, mask);
}

void catweasel_seek(catweasel_drive *d, int t)
{
    int x;

    if(t >= 80) {
	return;
    }

    x = t - d->track;
    if(!x)
	return;

    if(x >= 0) {
	CWSetCReg(d->contr, 1<<1, 0);
    } else {
	CWSetCReg(d->contr, 0, 1<<1);
	x = -x;
    }

    while(x--)
	CWTriggerStep(d->contr);

    d->track = t;
}

int catweasel_disk_changed(catweasel_drive *d)
{
    int ot, t, changed = 0;

    if(inb(d->contr->iobase+2) & 32) {
	if(!d->diskindrive) {
	    /* first usage of this drive, issue disk change */
	    d->diskindrive = 1;
	    changed = 1;
	} else {
	    /* there's still a disk in there, no change detected. */
	}
    } else {
	ot = d->track;
	if(ot == 79)
	    t = 78;
	else
	    t = ot + 1;
	
	catweasel_seek(d, t);
	catweasel_seek(d, ot);
	
	if(!(inb(d->contr->iobase+2) & 32)) {
	    if(!d->diskindrive) {
		/* drive still empty, nothing has happened */
	    } else {
		/* disk has been removed */
		d->diskindrive = 0;
		changed = 1;
	    }
	} else {
	    /* disk has been inserted */
	    d->diskindrive = 1;
	    changed = 1;
	}
    }

    return changed;
}

int catweasel_write_protected(catweasel_drive *d)
{
    return !(inb(d->contr->iobase+2) & 8);
}

int catweasel_read(catweasel_drive *d, int side, int clock, int time)
{
    int iobase = d->contr->iobase;

    if(!(inb(iobase+2) & 32))
	return 0;

    CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

    inb(iobase+1); /* ra reset */
    outb(clock*128, iobase+3);

    inb(iobase+1);
    inb(iobase+0);
    inb(iobase+0);
    outb(0, iobase+3); /* don't store index pulse */

    inb(iobase+1);

    inb(iobase+7); /* start reading */
    d->contr->msdelay(time); /* wait for one rotation */
    outb(0, iobase+1); /* stop reading, don't reset RAM pointer */

    outb(128, iobase+0); /* add data end mark */
    outb(128, iobase+0);

    inb(iobase+1); /* Reset RAM pointer */
    inb(iobase+0); /* ignore first byte */

    return 1;
}

static void
catweasel_wait_for_writing_flag (catweasel_drive *d,
				 int status,
				 int timeout_ms)
{
    int iobase = d->contr->iobase;
  
    while((inb(iobase+2) & 2) >> 1 != status && timeout_ms > 0) {
	d->contr->msdelay(20);
	timeout_ms -= 20;
    }
}

int
catweasel_write (catweasel_drive *d,
		 int side,
		 int clock,
		 int time)
{
    int iobase = d->contr->iobase;

    if(!(inb(iobase+2) & 32))
	return 0;

    CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

    inb(iobase+1);
    outb(clock*128, iobase+3);
    inb(iobase);
    outb(128, iobase+3);
    inb(iobase);
    inb(iobase);
    inb(iobase);
    inb(iobase);
    inb(iobase);
    inb(iobase);

    if(time > 0) {
	outb(0, iobase+5);
	d->contr->msdelay(time);
    } else {
	outb(0, iobase+7);
	catweasel_wait_for_writing_flag(d, 0, 300);
	d->contr->msdelay(190);
	catweasel_wait_for_writing_flag(d, 1, 50);
    }

    inb(iobase+1);
    return 1;
}
